#!/usr/bin/perl

### mkidentif V 0.3; copyright 2003 by Stefan Heimann
## input:
# 0 : normfile (infile)
# 1 : identiffile (main outfile)
# [2 : (optional)
#
#       "-prior=<scanner1>[, ...]" : list of scanner priorities,
#               for example: "-prior=fmw,fpr,vsc"; all scanners not listed have the same priority
#
#       "-workpath=<path>" : workpath for samples to identify full; to merge equal samples in different prepaths,
#               for example: "-workpath=ur-rest.021" : identify samples in ur-rest.021\* :
#                       this matches c:\heureka3\ur-rest.021 and q:\ur-rest.021
#               it is also like a view: samples not in -workpath will not worked up
#
#       "-tbs=<type 1>[, ...]" : list of testbeds to identify for
# ]

## priorities:
# 0: no priority (default) (see also input)
# 1: top priority; 2 : second priority; ...


if ($#ARGV < 1) {# proof number of given parameters
    print("ERROR: missing parameter!!! \n");
    exit 1;
}


## include modules ##

use FileHandle; # perl module for file handles
use normfile_methods; # methods for the NormFile, for instance read out the NormFile
use identiffile_methods; # methods for the IdentifFile, for instance read out the IdentifFile
use texttools; # tools for texthandling
use filehandling; # tools for file handling

## Global variables ##

$variant_seperator = "-"; # seperator for generic identification with multiple variants
$dropper_variant_mark = "_drv_"; # mark for extra dropper variant info
$infective_length_mark = "_"; # mark for infective length
$devolution_mark = "_d_"; # mark for (d)evolutions
$language_mark = "_l_"; # mark for languages
$addition_mark = "__"; # mark for additions
$extension_mark = "."; # mark for extensions

%normsamples = (); # samples and identifications from norm file; key: samplepath; value: identification-hash
@scanner_priorities = (); # array of scanner priorities : key: priority; value: scanner
%scanner_prior_set = (); # hash of scanner, which priority is set
%sample_identifications = (); # sample identifications hash; key: unified (relative) samplepath; value: single identification
$workpath = ""; # path to samples to identify
#@identify_testbeds; # list of testbeds to identify for

$NormFile = $ARGV[0]; # input-existing-normfile
$IdentifFile = $ARGV[1]; # old file of full sample identifications

if (defined($ARGV[2])) { # get options, if exist
        my $argvstr = "";
        for($zaehler = 0; $zaehler < @ARGV; $zaehler++) {
                $argvstr .= " ".$ARGV[$zaehler];
        }
        if ($argvstr =~ /-prior=(\S+)/i) {
                get_scanner_priorities($1); # get scanner priorities
        }
        if ($argvstr =~ /-workpath=(\S+)/i) {
                $workpath = uc($1); # set workpath
        }
        if ($argvstr =~ /-tbs=(\S+)/i) {
                @identify_testbeds = split(/\,/,uc($1)); # set list of testbeds to identify for
        }
}

$NewIdentifFile = $IdentifFile; # new file of full sample identifications
my @NormFiletmp1 = split(/\./,$NormFile); # temp var to split filename from extension
my @NormFiletmp2 = split(/_/,$NormFiletmp1[0]); # temp var to split "_norm" from filename, if exists
$DebugFile = $NormFiletmp2[0]."_mkidentif\.dbg"; # file for debugging
$NotNormFile = $NormFiletmp2[0]."_notnorm\.lst"; # list of out of norm sample identifications in normfile
$NotIdentifFile = $NormFiletmp2[0]."_notidentif\.lst"; # list of sample identifications, which could not be solved to a full identification
$RestFile = $NormFiletmp2[0]."_identifrest\.lst"; # list of sample identifications, which were not worked up, for instance because sample not in -workpath


## open / backup / create files ##

$DebugFH = new FileHandle "> $DebugFile" or do {
        print("ERROR creating $DebugFile : $!\n");
        closeall(); # close all opened files
        exit 1; # stop script
        };

$RestFH = new FileHandle "> $RestFile" or do {
        print("ERROR creating $RestFile : $!\n");
        closeall(); # close all opened files
        exit 1; # stop script
        };

$NotNormFH = new FileHandle "> $NotNormFile" or do {
        print("ERROR creating $NotNormFile : $!\n");
        closeall(); # close all opened files
        exit 1; # stop script
        };

$NotIdentifFH = new FileHandle "> $NotIdentifFile" or do {
        print("ERROR creating $NotIdentifFile : $!\n");
        closeall(); # close all opened files
        exit 1; # stop script
        };

$IdentifFile = &filehandling::backup_file($IdentifFile); # backup identiffile, if it exists
if ($IdentifFile ne "#new#") { # if an IdentifFile exists then read out full sample identifications
        %sample_identifications = &identiffile_methods::read_identiffile($IdentifFile);
}

$NewIdentifFH = new FileHandle "> $NewIdentifFile" or do {
        print("ERROR creating $NewIdentifFile : $!\n");
        closeall(); # close all opened files
        exit 1; # stop script
        };


## call here main-routines ##


print ($DebugFH "\$NormFile mkidentif : $NormFile\n");
print ($DebugFH "\$IdentifFile mkidentif : $IdentifFile\n");
print ($DebugFH "given scanner priorities :\n");
foreach $scanner (%scanner_priorities) {
        print ($DebugFH "  ".$scanner." :  ".$scanner_priorities{$scanner}."\n");
}

%normsamples = &normfile_methods::read_normfile($NormFile); # read out normfile

if ($workpath ne "") {
        %normsamples = merge_samples({%normsamples},$workpath); # merge equal samples scanned under different prepaths by using the given -workpath
}

identifiing(); # get full sample identifications out from the partial identifications

write_identif(); # write single sample identifications to NewIdentifFile

closeall();

&filehandling::rm_nullbytefiles($NotIdentifFile,$NotNormFile,$RestFile); # remove output files with null byte


### Sub-Routines ###


sub get_scanner_priorities { # set scanner priorities for given scanners

        my @scannerlist = split(/,/,$_[0]); # make an array out of the scannerlist-string

        $scanner_priorities[0] = ();

        #print ($DebugFH "\nget_scanner_priorities:\n\n");

        for ($counter = 0; $counter < @scannerlist; $counter++) {
                $scanner_priorities[$counter+1] = uc($scannerlist[$counter]);
                $scanner_prior_set{uc($scannerlist[$counter])} = "0";
        }
}


sub merge_samples { # merge equal samples scanned under different prepaths by using the given -workpath
#in: normsample hash; -workpath

        my %normsamples = %{$_[0]};
        my $workpath = $_[1];

        my %newnormsamples; # according to -workpath merged normsamples
        my $newsample; # new samplepath: cut of path before -workpath

                print ($DebugFH "\nmerge_samples:\n\n");

        foreach $sample (keys(%normsamples)) {
                #print ($DebugFH "sample: $sample\n");
                if ($sample =~ /\Q$workpath\E(\\\S+)/i) { # sample is in workpath
                        #print ($DebugFH "  sample is in workpath\n");
                        $newsample = $workpath.$1;
                        if (defined($newnormsamples{$newsample})) { # new samplepath does already exist => equal sample in different prepaths -> merge samples
                                #print ($DebugFH "    new samplepath: $newsample\n");
                                #print ($DebugFH "    new samplepath does already exist -> merge samples\n");
                                foreach $identification (keys(%{$normsamples{$sample}})) { # insert new identifications
                                        ${$newnormsamples{$newsample}}{$identification} = 0;
                                        #print ($DebugFH "      add identification: $identification\n");
                                }
                        }
                        else { # new samplepath does not exist -> copy identification hash into new sample hash
                                        #print ($DebugFH "     NEW samplepath does not exist -> copy identification hash into new sample hash\n");
                                $newnormsamples{$newsample} = $normsamples{$sample};
                                #foreach $identification (keys(%{$newnormsamples{$newsample}})) {
                                        #print ($DebugFH "      identification added: $identification\n");
                                #}
                        }
                }
                else { # sample is not in workpath -> Rest
                        print ($DebugFH "  sample is not in workpath -> Rest\n");
                        print ($RestFH "not in workpath: ".$sample."\n");
                        foreach $identification (keys(%{$normsamples{$sample}})) {
                                print ($RestFH "\t".$identification."\n");
                        }
                        print ($RestFH "\n"); # blankline for human readability
                }
        }

        return %newnormsamples;
}


sub identifiing {

        print($DebugFH "identifiing ...\n");

        my $sample; # path of a sample
        my $identification; # single, partial identification of a sample
        my %identifications; # hash of identifications; key: scanner; value: array of identificationparts
        my @identification_array; # array returned from to_identificationparts
        my $scanner; # scanner shortcut

        my @innorm_identifstrings; # save in norm identification strings to go arround : idstring -> idarray -> idstring , if an id is out of norm
        my $out_of_norm_before; # if more than one identification is out of norm, write sample only once to NoIdentifFile
        #my $identifcounter; # count total number of identifications
        my @noprior_identification; # merged identification of no priority scanner identifications

        foreach $sample (keys(%normsamples)) { # iterate all samples and get a full identification
                print ($DebugFH "sample : ".$sample."\n");
                if (not defined($sample_identifications{$sample})) { # proof, if single identification does already exist
                    ## initialize variables for this sample
                    %identifications = ();
                    @innorm_identifstrings = ();
                    $out_of_norm_before = 0;
                    foreach $identificationstr (keys(%{$normsamples{$sample}})) { # iterate all identificationstrings of this sample
                            @identification_array = to_identificationparts($identificationstr);
                            if (($identification_array[0] ne "") && ($identification_array[0] ne "NOTNORM")) { # identification partitioning was successfull
                                $scanner = $identification_array[0];
                                $identifications{$scanner} = $identification_array[1];
                                print ($DebugFH "  identif from $scanner added: ".$identifications{$scanner}."\n");

                                push(@innorm_identifstrings,$identificationstr);
                                #print ($DebugFH "Identification: ".$innorm_identifstrings[$#innorm_identifstrings]."\n");
                                #$identifcounter++;

                                if (not defined($scanner_prior_set{$scanner})) { # if no scanner priority was set for scanner -> set default priority
                                        push(@{$scanner_priorities[0]},$scanner); # append scanner to list of no priority scanners
                                        $scanner_prior_set{$scanner} = 0;
                                        print ($DebugFH "  no-priority-scanner added: $scanner\n");
                                }
                            }
                            else { # identification is out of norm
                                print ($DebugFH "  identification $identificationstr is out of norm\n");
                                if ($out_of_norm_before == 0) { # first identification, which is out of norm, for this sample
                                        print ($NotNormFH $sample."\n\tNOTNORM :  ".$identificationstr."\n");
                                        $out_of_norm_before = 1;
                                }
                                else { # write only the identification, if an other identification of this sample was out of norm before
                                        print ($NotNormFH "\tNOTNORM :  ".$identificationstr."\n");
                                }
                            }
                    }##end: foreach $identificationstr (keys(%{$normsamples{$sample}}))

                    if ($out_of_norm_before == 0) { # all identifications of this sample are in norm => make full identification

                            print ($DebugFH "  all identifications of this sample are in norm => make full identification\n");
                            $sample_identifications{$sample} = ["","","","","","","","","","","","",""]; # initialize full identification of sample
                            print ($DebugFH "  minimum priority: ".$#scanner_priorities."\n");
                            for($priority = 1; $priority < @scanner_priorities; $priority++) { # first, fill in informations per priorities
                                    if (defined($identifications{$scanner_priorities[$priority]})) { # if given priority scanner has made an identification
                                            print ($DebugFH "  proofed for priority $priority : $scanner_priorities[$priority] : ".$identifications{$scanner_priorities[$priority]}."\n");
                                            @{$sample_identifications{$sample}} = fillin_info($sample_identifications{$sample},$identifications{$scanner_priorities[$priority]});
                                            #for($index = 0; $index < @{$sample_identifications{$sample}}; $index++) { # iterate all identification parts
                                                    #if ($sample_identifications{$sample}[$index] eq "") { # if no information exists for this part => fill in information
                                                            #$sample_identifications{$sample}[$index] = $identifications{$scanner_priorities[$priority]}[$index];
                                                            #print ($DebugFH "    identification part at $index from $scanner with priority $priority added: ".$sample_identifications{$sample}[$index]."\n");
                                                    #}
                                            #}
                                    }
                            }##end: for($priority = 1; $priority < @scanner_priorities; $priority++)

                            @noprior_identification = compare_majority({%identifications}); # compare no priority scanner identifications per part and get an identification by identification majority

                            print ($DebugFH "    add identification parts with no priority : \n");
                            @{$sample_identifications{$sample}} = fillin_info($sample_identifications{$sample},\@noprior_identification);

                            @{$sample_identifications{$sample}} = complete_identification($sample_identifications{$sample});

                    }##end: if ($out_of_norm_before == 0)
                    else { # at least one identification was not in norm => write also in norm identifications to NotNorm
                            foreach $identificationstr (@innorm_identifstrings) {
                                    print ($NotNormFH "\tNORM :  ".$identificationstr."\n");
                            }
                    }
                }##end: if not defined($sample_identifications{$sample})
                else { # single sample identification already exists
                        print ($DebugFH "WARNING: single sample identification already exists:\n");
                        print ($DebugFH "  sample                : '".$sample."'\n");
                        print ($DebugFH "  single identification : ".to_identificationstring($sample_identifications{$sample})."\n");
                }
        }##end: foreach $sample (keys(%normsamples))
}##end: sub identifiing


sub fillin_info { # fill in missing informations
#in: current array of informations
#out: new array of informations

        @identif_parts = @{$_[0]}; # current identificaion part array
        @fill_identif_parts = @{$_[1]}; # identification parts to fill in

        my @proof_partinds = ("0","1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"); # partindices to proof

        my @new_identif_parts = @identif_parts; # new identification part array to return

        print ($DebugFH "\nfillin_info:\n\n");
        #print ($DebugFH "  \$#identif_parts: $#identif_parts\n");
        #for ($counter = 0; $counter <= $#identif_parts; $counter++) {
                #print ($DebugFH "  \$identif_parts\[$counter\]: '".$identif_parts[$counter]."'\n");
        #}

        my $counter = 0;
        while ( ($counter <= $#proof_partinds) && ($new_identif_parts[0] ne "NOTNORM") ) { # fill in new parts until all partindices proofed or an error occured
                $partindex = $proof_partinds[$counter]; # get next partindex to proof
                print ($DebugFH "  partindex to proof: $partindex\n");
                if (($partindex =~ /^(\d+)$/i) && ($identif_parts[$1] eq "")) { # empty part to fill found
                        $new_identif_parts[$1] = $fill_identif_parts[$1];
                        print ($DebugFH "  $1 <- '".$fill_identif_parts[$1]."'\n");
                }
                elsif ($partindex =~ /^(\d+)(x(\d+))+$/i) { # proof for given indices parts to fill in mutual exclusion
                        print ($DebugFH "    xor-proof : $partindex\n");
                        my $xor_partinds = $partindex; # xor part indices expression for work
                        my $info = 0; # flag for existing info
                        my $fillin_at = -1; # index to fill in new info; -1: index not found
                        while ($xor_partinds ne "") {
                                print ($DebugFH "  \$xor_partinds: ".$xor_partinds."\n");
                                if ($xor_partinds =~ /^(\d+)(x(.+))?$/i) {
                                        if ($3) { $xor_partinds = $3; } else { $xor_partinds = ""; } # reduce xor part indices expression

                                        if ( ($fill_identif_parts[$1] ne "") && ($fillin_at < 0) ) { # info to fill in found; only the first info counts
                                                $fillin_at = $1;
                                        }

                                        if ( ($info == 0) && ($identif_parts[$1] ne "") ) { # information exists
                                                $info = 1;
                                        }
                                        elsif (($info == 1) && ($identif_parts[$1] ne "")) { # information found second time => not in norm
                                                @new_identif_parts = ["NOTNORM"]; # indicate NOTNORM-ERROR
                                                $fillin_at = -1;
                                                $xor_partinds = "";
                                                print ($DebugFH "fillin_info: OUT OF NORM (xor) :  $1\n");
                                        }
                                }
                                else {
                                        print("      ERROR: \$xor_partinds: '".$xor_partinds."'\n");
                                        print("      ERROR: \$identif_parts[$1]: '".$identif_parts[$1]."'\n");
                                }
                        }##end: while ($xor_partinds ne "")
                        if ( ($info == 0) && ($fillin_at >= 0) ){ # no information was found => fill in info, if exists
                                $new_identif_parts[$fillin_at] = $fill_identif_parts[$fillin_at];
                                print ($DebugFH "  $fillin_at <- '".$fill_identif_parts[$fillin_at]."'\n");
                        }
                }##end: elsif "xor"

                $counter++;
        }##end: while ( ($counter <= $#proof_partinds) && ($new_identif_parts[0] ne "NOTNORM") )

        return @new_identif_parts;
}##end: fillin_info()


sub compare_majority { # compare identifications, insert identificationsparts, which are equal by the MAJORITY of the scanners
#in : identificationhash
#out : full identification array or []

        my %identifications = %{$_[0]};

        my @noprior_identifs = (); # identification array of no priority scanner identifications
        my @comp_identif = ("","","","","","","","","","","","",""); # compared identification, which will be returned to calling method

        print ($DebugFH "\ncompare_majority:\n\n");
        #print ("\ncompare_majority:\n\n");
        foreach $scanner (@{$scanner_priorities[0]}) {
                print ($DebugFH "  proof scanner: $scanner\n");
                #print ("  proof scanner: $scanner\n");
                if (defined ($identifications{$scanner})) {
                        push(@noprior_identifs,$identifications{$scanner}); # insert identification into no prior scanner identification list
                        print ($DebugFH "  insert identification into no prior scanner identification list: $scanner : ".$noprior_identifs[$#noprior_identifs]."\n");
                        #print ("  insert identification into no prior scanner identification list: $scanner : ".$noprior_identifs[$#noprior_identifs]."\n");
                        #foreach $identif_part (@{$noprior_identifs[$#noprior_identifs]}) {
                                #print ($DebugFH "  partcontent: $identif_part\n");
                        #}
                }
        }

        my %identifparts; # hash of identificationparts; key: identifpart; value : number of scanner with equal identification for this part
        my $majority_identif; # identification of the majority of the scanners for this part
        my $identifnumber; # number of no priority scanner identifications, which are not empty
        my $majority; # number of equal identifications must be greater than the absolute half of all noprior identifications for this part
        my $current_majority; # number of current majority
        my $identification;

        for($partindex = 0; $partindex <= $#comp_identif; $partindex++) { # make partial identification for all parts
                print ($DebugFH "  partindex: $partindex\n");
                ## initialize temporary variables for this part
                %identifparts = ();
                $majority_identif = "";
                $identifnumber = 0;
                $majority = 0;
                $current_majority = 0;
                #print ($DebugFH "  foreach \$identification (\@noprior_identifs)\n");
                foreach $identification (@noprior_identifs) { # count number of equal identifications in specific part
                        #print ($DebugFH "    if ('".${$identification}[$partindex]."' ne '')\n");
                        if (${$identification}[$partindex] ne "") { # not empty identification for this part
                                if (not defined($identifparts{${$identification}[$partindex]})) { # new identification for this part; exclude empty identifications ("")
                                        $identifparts{${$identification}[$partindex]} = 1;
                                }
                                else { # at least two scanners have got the same not empty identification for this part
                                        $identifparts{${$identification}[$partindex]}++;
                                }

                                $identifnumber++; # the number of not empty scanner identifications increased
                                $majority = int($identifnumber / 2) + 1; # the number to majority increased too
                                print ($DebugFH "  number = $identifnumber\n");
                                print ($DebugFH "  majority : $identifnumber >= $majority\n");
                                if ($identifparts{${$identification}[$partindex]} >= $majority) { # number of scanners is over the half => majority identification for this part is found
                                        print ($DebugFH "    majority identification for this part is found: ".$identifparts{${$identification}[$partindex]}."\n");
                                        $majority_identif = ${$identification}[$partindex]; # save current majority identification for this part
                                        $current_majority = $identifparts{${$identification}[$partindex]}; # save current majority number
                                }
                                elsif ($current_majority < $majority) { # current majority is not a majority -> reset
                                        print ($DebugFH "    current majority is not a majority -> reset : ".$current_majority." < ".$majority."\n");
                                        $current_majority = 0;
                                        $majority_identif = "";
                                }
                        } ##end: if ($identification[$partindex] ne "")
                }

                if ($majority_identif ne "") { # if majority identification for this part found, save it
                        $comp_identif[$partindex] = $majority_identif;
                }
        }

        return @comp_identif;
}


sub compare_unity { # compare identifications, insert identificationsparts, which are equal by ALL of the scanners
}


sub complete_identification { # complete missing parts of the identification
##in: identification array of identification parts, which might contain empty parts
##out: completed identification array

        my @identification = @{$_[0]}; # identification to complete

        my @complete; # array for information completing : key~partindex; value: hash : key: comparestring; value: array of informationparts

                # index:   0     1      2         3       4        5                 6           7         8         9          10          11    12
                # fields: [type, class, platform, family, variant, infective_length, devolution, language, addition, extension, filelength, crc1, crc2]
                ${$complete[2]}{"VBS"} =   ["script","","","","","","","","","vbs","","",""];
                ${$complete[2]}{"JS"} =    ["script","","","","","","","","","js","","",""];
                ${$complete[2]}{"JVS"} =   ["script","","","","","","","","","js","","",""];
                ${$complete[2]}{"IRC"} =   ["script","","","","","","","","","ini","","",""];
                ${$complete[2]}{"MIRC"} =  ["script","","","","","","","","","ini","","",""];
                ${$complete[2]}{"WBS"} =   ["script","","","","","","","","","","","",""]; # Word Basic Script
                ${$complete[2]}{"BAT"} =   ["exot","","","","","","","","","bat","","",""];
                ${$complete[2]}{"PERL"} =  ["exot","","","","","","","","","pl","","",""];
                ${$complete[2]}{"JV"} =    ["exot","","","","","","","","","cla","","",""];
                ${$complete[2]}{"WM"} =    ["macro","","","","","","","","","doc","","",""];
                ${$complete[2]}{"W2M"} =   ["macro","","","","","","","","","doc","","",""]; # Word2Macro
                ${$complete[2]}{"W97M"} =  ["macro","","","","","","","","","doc","","",""];
                ${$complete[2]}{"A97M"} =  ["macro","","","","","","","","","mdb","","",""];
                ${$complete[2]}{"O97M"} =  ["macro","","","","","","","","","","","",""];
                ${$complete[2]}{"P98M"} =  ["macro","","","","","","","","","mpp","","",""]; # ProjectPlan98Macro
                ${$complete[2]}{"X97M"} =  ["macro","","","","","","","","","xls","","",""];
                ${$complete[2]}{"XM"} =    ["macro","","","","","","","","","xls","","",""];
                ${$complete[2]}{"XF"} =    ["macro","","","","","","","","","xls","","",""]; # ExcelFormula
                ${$complete[2]}{"PP97M"} = ["macro","","","","","","","","","ppt","","",""];
                ${$complete[2]}{"APM"} =   ["macro","","","","","","","","","sam","","",""]; # ext: sam || smm # AmiProfessionalMacro
                ${$complete[2]}{"W32"} =   ["file","","","","","","","","","exe","","",""];
                ${$complete[2]}{"W16"} =   ["file","","","","","","","","","exe","","",""];
                ${$complete[2]}{"DOS"} =   ["file","","","","","","","","","exe","","",""];

                ${$complete[9]}{"VBS"} =   ["script","","vbs","","","","","","","","","",""];
                ${$complete[9]}{"JS"} =    ["script","","js","","","","","","","","","",""];
                ${$complete[9]}{"JVS"} =   ["script","","js","","","","","","","","","",""];
                ${$complete[9]}{"INI"} =   ["script","","irc","","","","","","","","","",""];
                ${$complete[9]}{"BAT"} =   ["exot","","bat","","","","","","","","","",""];
                ${$complete[9]}{"PL"} =    ["exot","","perl","","","","","","","","","",""];
                ${$complete[9]}{"CLA"} =   ["exot","","jv","","","","","","","","","",""];
                ${$complete[9]}{"DOC"} =   ["macro","","","","","","","","","","","",""];
                ${$complete[9]}{"DOT"} =   ["macro","","","","","","","","","","","",""];
                ${$complete[9]}{"XLS"} =   ["macro","","","","","","","","","","","",""];
                ${$complete[9]}{"MDB"} =   ["macro","","a97m","","","","","","","","","",""];
                ${$complete[9]}{"PPT"} =   ["macro","","pp97m","","","","","","","","","",""];
                ${$complete[9]}{"SAM"} =   ["macro","","apm","","","","","","","","","",""];
                ${$complete[9]}{"SMM"} =   ["macro","","apm","","","","","","","","","",""];
                ${$complete[9]}{"EXE"} =   ["file","","","","","","","","","","","",""];
                ${$complete[9]}{"COM"} =   ["file","","","","","","","","","","","",""];
                ${$complete[9]}{"SYS"} =   ["file","","","","","","","","","","","",""];
                ${$complete[9]}{"DLL"} =   ["file","","","","","","","","","","","",""];

             print ($DebugFH "\ncomplete_identification:\n\n");
             print ($DebugFH "\$#identification: ".$#identification."\n");
             for($counter = 0; $counter <= $#identification; $counter++) {
                print ($DebugFH "\$identification\[$counter\]: '".$identification[$counter]."'\n");
             }

             my $search_partindex; # current part index for matching strings
             my $replace_partindex; # current part index for informations to add in @complete, if no information on $replace_partindex
             print ($DebugFH "  for (\$search_partindex = 0; \$search_partindex <= $#complete; \$search_partindex++)\n");
             for ($search_partindex = 0; $search_partindex <= $#complete; $search_partindex++) { # try to match part in @complete with identification part
                #print ($DebugFH "    partindex1: $search_partindex\n");
                #print ($DebugFH "    \${\$complete\[$search_partindex\]}{\$identification\[$search_partindex\]}: ".${$complete[$search_partindex]}{$identification[$search_partindex]}."\n");
                #print ($DebugFH "    \$complete[\$search_partindex] :".$complete[$search_partindex]."\n");
                #print ($DebugFH "    \$identification\[$search_partindex\] :".$identification[$search_partindex]."\n");
                if (defined (${$complete[$search_partindex]}{$identification[$search_partindex]})) { # strings are matching
                                for ($replace_partindex = 0; $replace_partindex < @identification; $replace_partindex++) { # proof for adding missing informations for all parts
                                        #print ($DebugFH "      if ('".$identification[$replace_partindex]."' eq '')\n");
                                        if ($identification[$replace_partindex] eq "") { # missing part found
                                                        $identification[$replace_partindex] = uc(${$complete[$search_partindex]}{$identification[$search_partindex]}[$replace_partindex]); # get information from @complete, might also be empty
                                                        #print ($DebugFH "(1) missing part found: $replace_partindex <- '".uc(${$complete[$search_partindex]}{$identification[$search_partindex]}[$replace_partindex])."'\n");
                                                        print ($DebugFH "(2) missing part found: $replace_partindex <- '".$identification[$replace_partindex]."'\n");
                                        }
                                }
                }
        }

        return @identification;
}


sub to_identificationparts { # converts an identification string into an array
#in : identification string
#out : identification array

        my $identificationstr = $_[0]; # string to extract each identification part from

        my @identification_array; # array : [0]: $scanner|"NOTNORM" ; [1]: array of identification parts (without $scanner)

        my $type; # which type of malware : script, macro, (boot), file, exot
        my $class; # which class of malware : virus, trojan, worm, ...
        my $platform; # identified platform : w97m, w32, vbs, ...
        my $family; # identified family
        my $variant; # identified variant
        my $infective_length; # instad of a variant : length of infective code in byte
        my $devolution; # not a particular variant, but differend because of a programming error
        my $language; # regional language : de, fr, tw, ...
        my $addition; # like @MM
        my $scanner; # shortcut for the scanner, the identification belongs to
        my $extension; # extension for file format : exe, mdb, vbs, doc, ...
        my $filelength; # filelength of the sample (used by RFW)
        my $crc1; # first crc-checksum (from RFW)
        my $crc2; # second crc-checksum (from RFW)

        #print ($DebugFH "\nto_identificationparts:\n\n");
        #closures                  1        2       3       4       567          8                                9               10     11             12            13                     14     15             16     17                    18              19    20              21       (20+)23           24     25
        if ($identificationstr =~ /(\w+)?:\/(\w+)?\/(\w+)?\/(\S+)?\/((([a-z]{1,3}($variant_seperator[a-z]{1,3})*)?($devolution_mark(\d+))?($language_mark([a-z]{2}))?)|($infective_length_mark(\w+)))?($addition_mark(\w+))*($dropper_variant_mark([a-z]{1,3}))?\/(\w+)?($extension_mark(\w{1,3}))?\|((\d+\.)?\d+)?;(\w+)?;(\w+)?/i) {
                $type = uc($1);
                $class = uc($2);
                $platform = uc($3);
                $family = uc($4);
                $variant = uc($7);
                $infective_length = uc($14);
                $devolution = uc($10);
                $language = uc($12);
                $addition = uc($15);
                $addition .= uc($17);
                $scanner = uc($19);
                $extension = uc($21);
                $filelength = uc($22);
                $crc1 = uc($24);
                $crc2 = uc($25);

                $identification_array[0] = $scanner; # use while identifiing; while writing out: = "";
                $identification_array[1] = [$type,$class,$platform,$family,$variant,$infective_length,$devolution,$language,$addition,$extension,$filelength,$crc1,$crc2];
        }
        else { # identification is out of norm
                $identification_array[0] = "NOTNORM"; # indicate, that identification is out of norm
        }

        return @identification_array;
}


sub to_identificationstring { # converts an array of identification parts into an identification string
#in : identification array
#out : identification string

        # index:   0     1      2         3       4        5                 6           7         8         9          10          11    12
        # fields: [type, class, platform, family, variant, infective_length, devolution, language, addition, extension, filelength, crc1, crc2]
        my @identif_parts = @{$_[0]}; # identification array

        my $identification;

        print ($DebugFH "\nto_identificationstring:\n\n");
        #print ($DebugFH "  \$#identif_parts: $#identif_parts\n");
        #for ($counter = 0; $counter <= $#identif_parts; $counter++) {
                #print ($DebugFH "  \$identif_parts\[$counter\]: '".$identif_parts[$counter]."'\n");
        #}
        #                 type                    class                  platform               family
        $identification = $identif_parts[0].":\/".$identif_parts[1]."\/".$identif_parts[2]."\/".$identif_parts[3]."\/";
        #   variant
        if ($identif_parts[4] ne "") {
                $identification .= $identif_parts[4];
        }
        #      infective_length
        elsif ($identif_parts[5] ne "") {
                $identification .= $infective_length_mark.$identif_parts[5];
        }
        #   devolution
        if ($identif_parts[6] ne "") {
                $identification .= $devolution_mark.$identif_parts[6];
        }
        #   language
        if ($identif_parts[7] ne "") {
                $identification .= $language_mark.$identif_parts[7];
        }
        #   addition
        if ($identif_parts[8] ne "") {
                $identification .= $identif_parts[8];
        }
        #                                       extension             filelength             crc1                   crc2
        $identification .= "\/".$extension_mark.$identif_parts[9]."|".$identif_parts[10].";".$identif_parts[11].";".$identif_parts[12];

        $identification = uc($identification);

        return $identification;
}


sub write_identif {

        print ($DebugFH "write out into new IdentifFile : $NewIdentifFile\n");
        #print ("write out into new IdentifFile : $NewIdentifFile\n");

        #my $sample;
        foreach $sample (keys(%sample_identifications)) { # iterate all samples
                my $DestFH;
                print ($DebugFH "  \$sample_identifications{$sample}: '".$sample_identifications{$sample}."'\n");
                my $full_identified = proof_full_identif($sample_identifications{$sample});
                if ($full_identified == 1) { # sample is full identified
                        if (defined (@identify_testbeds)) { # wanted types or testbeds are spezified
                             my $in_testbed = 0; # flag for current sample identification is wanted or not
                             foreach $testbed (@identify_testbeds) { #
                                   if ($sample_identifications{$sample}[0] eq uc($testbed)) { # sample identification is wanted
                                      $DestFH = $NewIdentifFH; # write out to NewIdentif
                                      $in_testbed = 1;
                                      print ($::DebugFH "IN testbedlist: ".$sample."\t".to_identificationstring($sample_identifications{$sample})."\n");
                                   }
                             }
                             if ($in_testbed == 0) { # sample identification is not wanted
                                  print ($RestFH "Not in testbedlist:\n");
                                  $DestFH = $RestFH;
                                  print ($::DebugFH "NOT IN testbedlist: ".$sample."\t".to_identificationstring($sample_identifications{$sample})."\n");
                             }
                        }
                        else { # write out sample identifications for all types or testbeds
                             $DestFH = $NewIdentifFH;
                        }
                }
                elsif ($full_identified == -1) { # sample identification is out of norm
                        print ($NotNormFH "NOTNORM:\n");
                        $DestFH = $NotNormFH;
                }
                else { # sample is not full identified
                        $DestFH = $NotIdentifFH;
                }
                print ($DestFH $sample."\n");
                print ($DestFH "\t".to_identificationstring($sample_identifications{$sample})."\n");
                print ($DestFH "\n"); # blank line
        }
}

sub proof_full_identif { # proof, that full identification is made
#in: identification parts
#out: boolean for full identification

        my @identif_parts = @{$_[0]};

        my @essential_partinds = ("0", "1", "2", "3", "4", "9"); # essential part indices

        my $full_identified = 1; # boolean for full identification; 1 : full identificated

        print ($DebugFH "\nproof_full_identif:\n\n");
        #print ("proof_full_identif:\n");
        #print ($DebugFH "  \$#identif_parts: $#identif_parts\n");
        #for ($counter = 0; $counter <= $#identif_parts; $counter++) {
                #print ($DebugFH "  \$identif_parts\[$counter\]: '".$identif_parts[$counter]."'\n");
        #}

        full_identified: while (1) {
                foreach $partindex (@essential_partinds) { # iterate all essential part indices
                        print ($DebugFH "  partindex to proof: $partindex\n");
                        if (($partindex =~ /^(\d+)$/i) && ($identif_parts[$1] eq "")) { # proofed index is empty => not full identified
                                $full_identified = 0;
                                print ($DebugFH "    proof_full_identif: not full identified (and) :  $1\n");
                                last full_identified; # abort proofing of indices left
                        }
                        elsif ($partindex =~ /^(\d+)(x(\d+))+$/i) { # proof indices expression in mutual exclusion
                                #print ($DebugFH "    xor-proof : $partindex\n");
                                my $xor_partinds = $partindex; # indices expression for work
                                my $info = 0; # flag for info found; 0 : no info found
                                while ($xor_partinds ne "") { # iterate all indices in expression
                                        #print ($DebugFH "  \$xor_partinds: ".$xor_partinds."\n");
                                        if ($xor_partinds =~ /^(\d+)(x(.+))?$/i) { # expression not empty
                                                if ($3) { $xor_partinds = $3; } else { $xor_partinds = ""; } # reduce expression

                                                if ( ($info == 0) && ($identif_parts[$1] ne "") ) { # information found first time
                                                        $info = 1;
                                                }
                                                elsif (($info == 1) && ($identif_parts[$1] ne "")) { # information found second time => not in norm: only one information in one part index
                                                        $full_identified = -1; # indicate NOTNORM-ERROR
                                                        $xor_partinds = "";
                                                        print ($DebugFH "proof_full_identif: OUT OF NORM (xor) :  $1\n");
                                                        last full_identified; # abort proofing of indices left
                                                }
                                        }
                                        else {
                                             print("      ERROR: \$xor_partinds: '".$xor_partinds."'\n");
                                             print("      ERROR: \$identif_parts[$1]: '".$identif_parts[$1]."'\n");
                                        }
                                }##end: while ($xor_partinds ne "")
                                if ($info == 0) { # no information was found => not full identified
                                        $full_identified = 0;
                                        print ($DebugFH "proof_full_identif: not full identified (xor) :  $partindex\n");
                                        last full_identified; # abort proofing of indices left
                                }
                        }##end: elsif "xor"
                }##end: foreach $partindex (@essential_partfinds)

                last full_identified; # leave "while true"
        }##end: while "true"

        return $full_identified;
}


sub closeall { # close all files
    close($DebugFH);
    close($NormFH);
    if ($IdentifFH) { close($IdentifFH); }
    close($NewIdentifFH);
    close($NotIdentifFH);
    close($RestFH);
    close($NotNormFH);
}